// Solution pour Question 2(d)

// "Tournoi de foot", on utilise les solutions pour Questions 2(a-c)

#include <stdlib.h>
#include <assert.h>
#include <semaphore.h>
#include <stdio.h>
#include <pthread.h>
#include <math.h>
#include <time.h>
#include <stdarg.h>

#define N 16

long int x;

sem_t tour[N-1];	// pour assurer l'ordre de matchs
sem_t mutex;		// pour accès écriture exclusif
sem_t sync[2];		// pour synchronizer les deux joueurs d'un match

pthread_t t[N];


// Émettre un message sur l'écran. Un sémaphore supplémentaire assure
// l'accès exclusif à l'écran, uniquement pour des raisons cosmétiques.
sem_t output;

void message (int id, char* msg, ...)
{
	va_list args;
	sem_wait(&output);
	printf("[id %d] ",id);
	va_start(args,msg);
	vprintf(msg,args);
	va_end(args);
	puts("");
	sem_post(&output);
}


void* f (void* arg)
{
	// extraire les deux paramètres
	long int id = ((long int)arg) / N;
	long int score = ((long int)arg) % N;

	// on joue les matches dans l'ordre de (c)
	int nr = id, base = 0, plus = N/2;
	while (base < N-1)
	{
		int match = base+nr/2;
		int joueur = nr & 1;
		int autre = 1-joueur;
		int score_autre;

		// attendre le match (on libère deux créneaux par match)
		sem_wait(tour+match);

		message(id,"je joue au match no %d",match);

		// le protocôle de (a) pour échanger les scores
		if (joueur) x = score;
		sem_post(sync+joueur); sem_wait(sync+autre);
		if (!joueur) { score_autre = x; x = score; }
		sem_post(sync+joueur); sem_wait(sync+autre);
		if (joueur) score_autre = x;

		// gagné ou perdu ? Attention: synchronisation supplémentaire
		// pour éviter interférence entre matchs.
		message(id,"le score est %d - %d",score,score_autre);
		if (score < score_autre) { message(id,"j'ai perdu");
				  sem_post(sync+joueur); break; }
		sem_wait(sync+autre);
		if (match == N-2) { message(id,"je gagne le tournoi"); break; }

		// on libère le match suivant
		sem_post(tour+match+1);
		sem_post(tour+match+1);

		base += plus; nr = nr/2; plus = plus/2;
	}

	return NULL;
} 

// créer une permutation de 0..N-1
void make_perm (long int* p)
{
	int i;
	for (i = 0; i < N; i++) p[i] = i;
	for (i = 0; i < N; i++)
	{
		int j = floor(drand48()*16);
		int tmp = p[j]; p[j] = p[i]; p[i] = tmp;
	}
}

int main(int argc,char **argv)
{
	pthread_t t[N];
	long int i, ids[N], scores[N];

	srand48(time(NULL));
	make_perm(ids);
	make_perm(scores);

	sem_init(tour+0,0,2);	// libérer match 0, bloquer les autres
	for (i = 1; i < N-1; i++) sem_init(tour+i,0,0);
	sem_init(sync+0,0,0); sem_init(sync+1,0,0);
	sem_init(&mutex,0,1);
	sem_init(&output,0,1);

	for (i = 0; i < N; i++) pthread_create(t+i,NULL,f,
					(void*)(ids[i]*N+scores[i]));
	for (i = 0; i < N; i++) pthread_join(t[i],NULL);
}
